import os


def process_readme(readme):
    """ """
    sections = readme.split("\n#")

    sections_processed = []
    for section in sections:
        sub_sections = section.split("\n")
        section_title = sub_sections[0].strip()
        section_body = sub_sections[1:]
        section_body = "\n".join(section_body).strip()
        code_snippet = section_body.split("``") if "```" in section_body else None
        out_snippet = None
        if code_snippet:
            for i, part in enumerate(code_snippet):
                if part.startswith("`Python"):
                    out_snippet = f"""``{"".join(code_snippet[i:i+1])}```"""
                    break
        sections_processed.append(
            (section_title, section_body, out_snippet, sub_sections, section)
        )

    return sections_processed


def make_readme():
    with open("build_tools/README.template.md", "r") as f:
        main_readme = f.read()

    main_folder = "streamlitextras"
    module_folders = [
        f
        for f in os.listdir(main_folder)
        if os.path.isdir(os.path.join(main_folder, f))
    ]
    for module_folder in module_folders:
        module_path = os.path.join(str(main_folder), module_folder)
        module_readme_file = os.path.join(module_path, "README.md")
        if not os.path.exists(module_readme_file):
            continue
        with open(module_readme_file, "r") as f:
            readme = f.read()
        sections = process_readme(readme)
        inserted_title = sections[0][0].replace("# Streamlit Extras ", "")
        inserted_description = sections[0][0]
        inserted_usage = sections[1][1]
        inserted_usage = sections[1][2]

        title_match = f"@@{inserted_title.upper()}".replace(" ", "").strip()
        footer = f"\n\nSee the [package readme]({module_path}) or [API docs](https://streamlitextras.readthedocs.io/en/latest/api/streamlitextras.html) for more details.\n"
        main_readme = main_readme.replace(title_match, inserted_usage + footer)

    with open("README.md", "w") as f:
        f.write(main_readme)

    return main_readme


if __name__ == "__main__":
    make_readme()



from collections.abc import Callable, Iterable, Mapping
import os
import time
import threading
from streamlit.runtime.scriptrunner import (
    add_script_run_ctx,
    get_script_run_ctx,
    RerunException,
    ScriptRunContext,
)
from typing import Any, Callable, Optional

# script_dir = os.path.dirname(os.path.realpath(__file__))
script_dir = os.getcwd()

default_id = "default"


def trigger_file_path(unique_id: str = default_id) -> str:
    trigger_file_path = os.path.join(script_dir, f"reruntrigger_{unique_id}.py")
    if not os.path.exists(trigger_file_path):
        with open(trigger_file_path, "w") as f:
            f.write(f"timestamp = {time.time()}")
    return trigger_file_path


lock = threading.Lock()


def last_trigger_time(unique_id: str = default_id) -> int:
    """
    Returns the seconds since last writing the trigger file
    """
    modified_time = os.path.getmtime(trigger_file_path(unique_id))
    modified_time_seconds = time.time() - modified_time
    return modified_time_seconds


def trigger_rerun(
    unique_id: str = default_id, last_write_margin: int = 1, delay: int = 0
) -> None:
    """
    Triggers treamlit to rerun the current page state.
    runOnSave must be set to true in config.toml

    :param str unique_id: Unique ID to be triggered, should be set per session e.g. user id or a hash you create in their session state.
    :param int last_write_margin:
        If the file was modified less than this many seconds ago, the rerun will not be performed
    :param int delay: sleep for this many seconds before writing the rerun trigger
    """
    if delay:
        time.sleep(delay)
    with lock:
        modified_time_seconds = last_trigger_time(unique_id)
        if last_write_margin == 0 or modified_time_seconds > last_write_margin:
            trigger_file = trigger_file_path(unique_id)
            print("Writing trigger", trigger_file, flush=True)
            with open(trigger_file, "w") as f:
                f.write(f"timestamp = {time.time()}")
    # https://github.com/streamlit/streamlit/issues/1792
    # https://discuss.streamlit.io/t/using-streamlit-with-multithreading/30990
    # https://discuss.streamlit.io/t/how-to-run-a-subprocess-programs-using-thread-inside-streamlit/2440/2
    # https://discuss.streamlit.io/t/how-to-monitor-the-filesystem-and-have-streamlit-updated-when-some-files-are-modified/822/3


def thread_wrapper(
    thread_func,
    rerun_st=True,
    last_write_margin: int = 1,
    delay: int = 0,
    trigger_unique_id: str = default_id,
    *args,
    **kwargs,
) -> None:
    """
    Wrapper for running thread functions
    For parameters see streamlit_thread() and trigger_rerun()
    """
    # print("Hashseed in thread:", os.environ.get("PYTHONHASHSEED", False))
    thread_func(*args, **kwargs)
    if rerun_st is True:
        trigger_rerun(trigger_unique_id, last_write_margin, delay)


def streamlit_thread(
    thread_func: Callable,
    args: tuple = (),
    kwargs: dict = {},
    rerun_st: bool = True,
    last_write_margin: int = 1,
    delay: int = 0,
    script_run_context: ScriptRunContext | None = None,
    autostart: bool = True,
    trigger_unique_id: str = default_id,
    error_handler: Callable | None = None,
) -> str:
    """
    Spawns and starts a threading.Thread that runs thread_func with the passed args and kwargs

    :param Callable thread_func: The function to run in the thread
    :param tuple args: The args to pass to the function in the thread
    :param dict kwargs: The kwargs to pass to the function in the thread
    :param bool rerun_st: Whether to rerun streamlit after the thread function finishes

    :param Callable error_handler: Error handler function that takes the thread exception as an argument

    :returns: The name of the thread. Can use get_thread to get the threading.Thread instance
    """
    # print("Thread entry hashseed:", os.environ.get("PYTHONHASHSEED", False))
    args = (thread_func, rerun_st, last_write_margin, delay, trigger_unique_id, *args)
    thread = PropagatingThread(
        target=thread_wrapper, error_handler=error_handler, args=args, kwargs=kwargs
    )

    if not script_run_context:
        script_run_context = get_script_run_ctx()
    add_script_run_ctx(thread, script_run_context)

    time.sleep(0.4)
    if autostart is True:
        thread.start()

    return thread.name


def get_thread(thread_name) -> Optional[threading.Thread]:
    """
    Gets the threading.Thread instance thats name attribute matches thread_name

    :param thread_name: The name attribute of the thread to look for.

    :returns: The threading.Thread or None if theres no thread with the supplied thread_name
    """
    threads = threading.enumerate()
    target_thread = None
    for thread in threads:
        if thread.name == thread_name:
            target_thread = thread
            break
    return target_thread


class PropagatingThread(threading.Thread):
    def __init__(self, *args, **kwargs) -> None:
        self.error_handler = kwargs.get("error_handler", None)
        del kwargs["error_handler"]
        super().__init__(*args, **kwargs)

    def run(self):
        self.exc = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except RerunException as e:
            self.exc = e
        except BaseException as e:
            self.exc = e
            if self.error_handler and callable(self.error_handler):
                self.error_handler(e)
            else:
                raise

    def join(self, timeout=None):
        super(PropagatingThread, self).join(timeout)
        if self.exc:
            if self.error_handler and callable(self.error_handler):
                self.error_handler(self.exc)
            else:
                raise self.exc
            # raise RuntimeError('Exception in thread') from self.exc
        return self.ret



import base64
from io import BytesIO
from requests import get
from datetime import datetime

import streamlit as st
from streamlit.runtime.uploaded_file_manager import UploadedFile


class CustomHTML:
    """
    Class containing custom HTML/CSS to be injected into the streamlit app
    """

    @property
    def st_info_classes(self) -> list:
        """
        Classes used by st.info(), used for creating a custom info box with self.custom_el()
        """
        return ["st-at", "st-ax", "st-aw", "st-av", "st-au"]

    @property
    def adjust_st_style(self) -> str:
        """
        Hides the streamlit footer and menus and makes other adjustments
        """
        code = """
            <style>
            #MainMenu {display: none !important;}
            footer {display: none !important;}
            header {display: none !important;}
            div[class*="viewerBadge"] {display: none !important;}
            .block-container {padding-top: 0px;}

            div.stButton > button:first-child {
                background-color: #0E4176;
                color: #ffffff;
            }
            div.stButton > button:hover {
                background-color: #0E4176;
                color: #ff0000;
            }
            div.stButton > button:disabled, button[disabled] {
                background-color: #F0F2F6;
                color: grey;
            }
            p.a {
                font: bold 12px Courier;
            }
            </style>
            """
        return code

    @staticmethod
    def centered_image(image_path: str) -> None:
        """
        Creates 3 streamlit columns with a centered image

        :param str image_path: The URL or local path of an image
        This is currently unused.
        """

        col1, col2, col3 = st.columns(3)

        with col1:
            st.write(" ")

        with col2:
            if image_path:
                st.image(image_path, use_column_width="always")
            else:
                st.image("https://static.streamlit.io/examples/dog.jpg")

        with col3:
            st.write(" ")

        st.image(image_path, use_column_width="always")

    @staticmethod
    def anchor_el(anchor_id: str):
        """
        Creates a div to be used as a # anchor

        :param str anchor_id: the string to set the id property to on the div element
        """
        return f"<div id='{anchor_id}' style='content-visibility: hidden; padding: 0px; margin: 0px;'>{anchor_id}</div>"

    @staticmethod
    def bytes_to_data_uri(
        byteslike_object: BytesIO | UploadedFile | bytes,
        mime_type: str | None = None,
    ) -> str:
        """
        Creates a data URI from a bytesIO object

        :param Union[BytesIO, UploadedFile] byteslike_object: BytesIO or bytes or any class that inherits them
        :param str mime_type: The mimetype to set on the data URI
        :returns: The data URI as a string
        """
        data = None
        try:
            data = byteslike_object.getvalue()
        except:  # noqa: E722
            data = byteslike_object
        if not mime_type:
            mime_type = "application/octet-stream"
        uri = f"data:{mime_type};base64,{base64.b64encode(data).decode()}"
        return uri

    @classmethod
    def download_link(
        cls,
        byteslike_object: BytesIO | UploadedFile | bytes,
        filename: str,
        link_text: str = "",
        mime_type: str | None = None,
    ) -> str:
        """
        Creates a data URI from a bytesIO object

        :param Union[BytesIO, UploadedFile] byteslike_object: BytesIO or bytes or any class that inherits them
        :param str filename: The filename presented for the download file
        :param str link_text: The text to put inside the link element, defaults to empty string
        :param Optional[str] mime_type: The mimetype to set on the data URI. Defaults to "application/octet-stream"
        """
        link_props = {
            "download": filename,
            "href": cls.bytes_to_data_uri(byteslike_object),
        }
        download_link = cls.custom_el(
            link_text, "a", custom_properties=link_props, classes=["auto-download"]
        )
        return download_link

    @staticmethod
    def custom_el(
        text: str,
        tag: str = "div",
        align: str = "",
        color: str = "",
        anchor: str = "",
        classes: list = [],
        extra_style: str = "",
        props: dict = {},
    ) -> str:
        """
        Creates HTML string for a custom element based on the function parameters.
        There is no validation so ensure calls to this function are correct.

        :param str text: the innerText/innerHTML of the element
        :param str tag: the HTML tag of the element
        :param str align: the CSS text-align property to apply to the element style
        :param str color: the css color property to apply to the element style
        :param str anchor: if provided, will put an empty div element with this value as an id next to the main element. used for creating anchors.
        :param list classes: list of css classes to apply to the class property of the element
        :param str extra_style: string of properly formatted CSS rules to be applied to the element style tag
        :param dict props: dictionary of properties and values to apply to the element in format {"property": "value"}
        """

        classes_string = ""
        for c in classes:
            classes_string += f"{c} "

        style_string = ""
        if align and align != "":
            style_string += f"text-align: {align};"
        if color and color != "":
            style_string += f"color: {color};"
        style_string += extra_style

        properties_string = ""
        for name, value in props.items():
            properties_string += f' {name}="{value}"'

        code = f"""<{tag} style="{style_string}" class="{classes_string}"{properties_string}>{text}</{tag}>"""
        if anchor != "":
            code = (
                f"<div id='{anchor}' style='padding: 0px; margin: 0px;'></div>" + code
            )
        return code

    @staticmethod
    def display_speaker(speaker_id: str) -> str:
        """
        Creates HTML text for a simple paragraph element with the innerText as speaker_id

        :param str speaker_id: The text to be inserted in the element innerText/innerHTML
        """
        code = f"""<p class="a">{speaker_id}</p>\n"""

        return code

    @staticmethod
    def display_text(text: str) -> str:
        """
        Creates HTML text for a styled paragraph element with the innerText as text

        :param str text: The text to be inserted into the paragraph innerText/innerHTML
        """
        code = f"""<p style="font-size: medium; font-family: sans-serif">{text}</p>\n"""
        return code

    @staticmethod
    def display_table(table_data: list[list], style: str = "") -> str:
        """
        Creates HTML table based on table_data which is a 2D list
        Each list item is a set of columns. The first item is the header.

        :param list[list] table_data: The data to fill the table in
        :param str style: The CSS style tag to apply to the table
        """
        table = f'<table style="{style}">\n'
        for i, row in enumerate(table_data):
            tag = "<th>" if i == 0 else "<td>"
            table += "<tr>\n"
            for column in row:
                table += f"""{tag}{column}{tag.replace("<", "</")}\n"""
            table += "</tr>\n"
        table += "</table>"

        return table

    @staticmethod
    def display_audio(
        audio_file_url: str,
        timestamp: str | int = "",
        id: str = "audio",
        classes: list = [],
    ) -> str:
        """
        Creates a HTML audio element with a specified id and special class,
        this enables more advanced handling of the elements than st.audio()

        :param str audio_file_url: The url to the audio file for the element
        :param str Union[str, int]: The timestamp in the format HH:MM:SS to apply to the URL fragment, alternatively can be the number of seconds
        :param str id: The id property of the audio element
        :param list classes: list of css classes to apply to the class property of the element
        """
        classes_string = "stAudio customAudio"
        for c in classes:
            classes_string += f" {c}"

        audio = f"""
        <audio id="{id}" controls preload
            src="{audio_file_url}{"#t="+str(timestamp) if timestamp != "" and timestamp != 0 else ""}"
            class="{classes_string}"
            style="width: 100%;"
            oncanplaythrough="">
        </audio>
        """

        return audio

    lottie_url = (
        "https://lottie.host/250d8ce6-e7eb-43a7-ab4f-1d78edc49b9d/2TLv74crQZ.json"
    )

    @classmethod
    def load_lottieurl(
        cls,
        lottie_url: str | None = None,
    ) -> dict | None:
        """
        Loads and returns JSON from the lottie library
        This is used for animations
        """
        if not lottie_url:
            lottie_url = cls.lottie_url
        r = get(lottie_url)
        if r.status_code != 200:
            return None
        return r.json()

    @property
    def lottie_data(self):
        return self.load_lottieurl()


custom_html = CustomHTML()
audio_extensions = ["PCM", "WMA", "MP4", "M4A", "WAV", "AIFF", "MP3", "AAC"]


def readable_datestr(target_datetime: datetime, format: str = "%d/%m/%Y"):
    return target_datetime.strftime(format)



# File contains utility or helper functions
import os
from typing import Optional
from streamlit.runtime.uploaded_file_manager import UploadedFile


def repr_(
    cls, ignore_keys: Optional[list[str]] = None, only_keys: Optional[list[str]] = None
) -> str:
    """
    Returns a string detailing a class attributes from cls.__dict__
    Makes nice printing for __repr__ implementations

    :param ignore_keys: If provided, these keys will be not be included in the attribute string
    :param only_keys: If provided, these keys will be the only ones included in the attribute string
    """
    if not hasattr(cls, "__dict__"):
        return repr(cls)
    if not ignore_keys or only_keys:
        ignore_keys = []
    classname = cls.__class__.__name__

    try:
        args = ", ".join(
            [
                f"{k}={repr(v)}"
                for (k, v) in cls.__dict__.items()
                if k not in ignore_keys and (k in only_keys if only_keys else True)
            ]
        )
    except RecursionError:
        args = "Too much recursion"
    return f"{classname}({hex(id(cls))}, {args})"


def save_file(st_file_object: UploadedFile, to_path: str) -> str:
    """
    Saves a streamlit UploadedFile BytesIO object to the given relative path

    :param UploadedFile st_file_object: The UploadedFile bytes object to save to disk - contains filename and metadata
    :param str to_path: The relative path to a folder to save the file to
    :returns str: Will return the relative path which can be used as a URL
    """
    os.makedirs(to_path, exist_ok=True)
    path = os.path.join(to_path, st_file_object.name)
    with open(path, "wb") as f:
        f.write(st_file_object.getbuffer())
    return "/" + path


def where_stack():
    import inspect

    stack = inspect.stack()
    the_class = stack[1][0].f_locals["self"].__class__.__name__
    the_method = stack[1][0].f_code.co_name

    print("I was called by {}.{}()".format(the_class, the_method))



from streamlitextras.logger import log
from streamlitextras.utils import repr_
from streamlitextras.authenticator.utils import handle_firebase_action
from streamlitextras.authenticator.exceptions import (
    AuthException,
    LoginError,
    RegisterError,
    ResetError,
    UpdateError,
)


class User:
    """
    This class is used as an interface for Authenticators users
    """

    def __init__(self, authenticator, auth_data, login_data={}, debug: bool = False):
        """
        Initializes a user account with associated firebase tokens and account information

        :param Authenticator authenticator: The associated Authenticator class that spawned this user
        :param dict auth_data:
            Dict containing session_cookie, decoded_claims and user_record
        :param dict login_data:
            The data from the initial login, not always provided (when reading session cookies)
        """
        self.debug = debug

        self.authenticator = authenticator
        self.auth_data = auth_data
        self.login_data = login_data

        self.idToken = login_data.get("idToken", None)
        self.expiresIn = login_data.get("expiresIn", None)
        self.registered = login_data.get("registered", None)
        self.refreshToken = login_data.get("refreshToken", None)

        self.user_record = auth_data["user_record"]
        self.session_cookie = auth_data["session_cookie"]
        self.decoded_claims = auth_data["decoded_claims"]

        self.uid = self.decoded_claims["uid"]
        self.localId = self.user_record["localId"]

        self.email = self.user_record["email"]
        self.emailVerified = self.user_record["emailVerified"]
        self.displayName = self.user_record.get("displayName", None)
        self.createdAt = self.user_record.get("createdAt", None)
        self.lastLoginAt = self.user_record.get("lastLoginAt", None)
        self.lastRefreshAt = self.user_record.get("lastRefreshAt", None)
        self.passwordUpdatedAt = self.user_record.get("passwordUpdatedAt", None)
        self.providerUserInfo = self.user_record.get("providerUserInfo", None)
        self.validSince = self.user_record.get("validSince", None)
        self.photoUrl = self.user_record.get("photoUrl", None)

        self.account_info = login_data.get("account_info", None)
        self.users = self.account_info["users"] if self.account_info else None
        self.user = self.users[0] if self.users else None
        self.disabled = self.user.get("disabled", None) if self.user else None
        self.customAuth = self.user.get("customAuth", None) if self.user else None

    def refresh_token(self):
        refreshed = None
        refresh_errors = {
            "TOKEN_EXPIRED": "Too many recent sessions. Please try again later."
        }
        user_refresh, refresh_error = handle_firebase_action(
            self.authenticator.auth.refresh,
            LoginError,
            refresh_errors,
            self.refreshToken,
        )
        if not refresh_error:
            self.login_data = {**self.login_data, **user_refresh}
            for key in self.__dict__.keys():
                if key in user_refresh and hasattr(self, key):
                    setattr(self, key, user_refresh[key])
            self.authenticator._create_session(self.login_data)
            refreshed = True
            if self.debug:
                log.info(f"Users tokens refreshed {self.uid}.")
        else:
            log.error(f"Error refreshing users firebase token {refresh_error}")
            refreshed = False

        return refreshed

    @property
    def firebase_user(self):
        """
        Gets the UserRecord object from the official firebase python SDK

        # https://firebase.google.com/docs/reference/admin/python/firebase_admin.auth#firebase_admin.auth.UserRecord
        """
        self.authenticator.service_auth.get_user(self.uid)

    @property
    def is_admin(self):
        """
        Returns true if users firebase id is in self.authenticator.admin_ids
        """
        admin_ids = self.authenticator.admin_ids
        return self.localId in admin_ids

    @property
    def is_developer(self):
        """
        Returns true if users firebase id is in self.authenticator.developer_ids
        """
        developer_ids = self.authenticator.developer_ids
        return self.localId in developer_ids

    def __repr__(self) -> str:
        return repr_(
            self,
            ["passwordHash", "login_data", "refreshToken", "idToken", "authenticator"],
            only_keys=["localId", "email"],
        )



import jwt
import pytz
import time
import dateutil.parser
from types import ModuleType
from typing import Union, Optional, Tuple, TypeVar, Type, Callable
from datetime import datetime, timedelta
from http.cookies import SimpleCookie

import streamlit as st
from streamlit.web.server.websocket_headers import _get_websocket_headers
from streamlit.delta_generator import DeltaGenerator
from streamlitextras.logger import log
from streamlitextras.utils import repr_
from streamlitextras.authenticator.user import User
from streamlitextras.authenticator.exceptions import (
    AuthException,
    LoginError,
    RegisterError,
    ResetError,
    UpdateError,
)
from streamlitextras.authenticator.utils import handle_firebase_action
from streamlitextras.webutils import get_user_timezone
from streamlitextras.cookiemanager import get_cookie_manager

import json
import requests
import pyrebase
import firebase_admin
from firebase_admin import auth as service_auth
from streamlitextras.helpers import custom_html

config = dict(st.secrets["firebase"])
pyrebase_service = pyrebase.initialize_app(config)
auth = pyrebase_service.auth()
db = pyrebase_service.database()
storage = pyrebase_service.storage()


# PyPI doesn't contain the latest git commits for pyrebase4
def update_profile(
    id_token, display_name=None, photo_url=None, delete_attribute=None, *args
):
    """
    Pyrebase on PyPI seems to be missing latest commits - adding this in here.
    https://firebase.google.com/docs/reference/rest/auth#section-update-profile
    """
    request_ref = (
        "https://identitytoolkit.googleapis.com/v1/accounts:update?key={0}".format(
            auth.api_key
        )
    )
    headers = {"content-type": "application/json; charset=UTF-8"}
    data = json.dumps(
        {
            "idToken": id_token,
            "displayName": display_name,
            "photoURL": photo_url,
            "deleteAttribute": delete_attribute,
            "returnSecureToken": True,
        }
    )
    request_object = requests.post(request_ref, headers=headers, data=data)
    pyrebase.pyrebase.raise_detailed_error(request_object)
    return request_object.json()


auth.update_profile = update_profile

UserInherited = TypeVar("UserInherited", bound="User", covariant=True)


class Authenticator:
    """
    Authenticator is used to handle firebase authentication,
    as well as create streamlit widgets for login, registration,
    and other account operations.

    :param str authenticator_name: named key for this Authenticator class, used to store it in session state. Used for tracking without classvar reference.
    :param str cookie_name: the name of the auth cookie to be saved in the browser
    :param str cookie_key: the key used to encrypt the JWT stored in the auth cookie
    :param str session_name: the name of the key to store in st.session_state and the JWT token (with exp_date), containing the decoded auth cookie to
    :param UserInherited user_class: Optional inherited User class to store the User state in
    :param list admins_ids:
        List of firebase user ids (localId) used for checking User.is_admin
        Can also be | seperated string that will be parsed to a list - useful for parsing from a secrets config file
    :param list developer_ids:
        List of firebase user ids (localId) used for checking User.is_developer
        Can also be | seperated string that will be parsed to a list - useful for parsing from a secrets config file
    :param bool require_email_verification: If set to True (the default), users will not be able to log in until they verify their email address
    """

    def __init__(
        self,
        cookie_name: str,
        cookie_key: str,
        session_expiry_seconds: int = 3600 * 24 * 12,
        session_name: str = "user",
        authenticator_name: str = "authenticator",
        user_class: Optional[Type[UserInherited]] = None,
        admin_ids: Optional[Union[list, str]] = None,
        developer_ids: Optional[Union[list, str]] = None,
        require_email_verification: bool = True,
        attempt_fallback_cookies: bool = False,
        debug: bool = False,
    ):
        self.debug = debug

        if authenticator_name == session_name:
            raise ValueError(
                "authenticator_name can't be the same as session_name (st.session_state conflict)"
            )

        if not user_class:
            user_class = User

        self.cookie_name = cookie_name
        self.session_name = session_name
        self.session_expiry_seconds = session_expiry_seconds
        self.authenticator_name = authenticator_name
        self.cookie_key = cookie_key
        self.last_user = None
        self.user_class = user_class
        self.require_email_verification = require_email_verification
        self.attempt_fallback_cookies = attempt_fallback_cookies

        self.storage = storage
        self.auth = auth
        self.db = db

        self.logged_out = None

        if not admin_ids:
            admin_ids = []
        if type(admin_ids) not in [str, list]:
            raise ValueError("admin_ids must be a list or a | seperated string")
        self.admin_ids = admin_ids.split("|") if type(admin_ids) == str else admin_ids

        if not developer_ids:
            developer_ids = []
        if type(developer_ids) not in [str, list]:
            raise ValueError("developer_ids must be a list or a | seperated string")
        self.developer_ids = (
            developer_ids.split("|") if type(developer_ids) == str else developer_ids
        )

        # Streamlit refreshes a lot recreating this class - keep the same form between refreshes
        self.current_form = None
        if (
            self.authenticator_name in st.session_state
            and st.session_state[self.authenticator_name]
            and st.session_state[self.authenticator_name].current_form
        ):
            self.current_form = st.session_state[self.authenticator_name].current_form

        self.service_credentials = firebase_admin.credentials.Certificate(
            dict(st.secrets["gcp_service_account"])
        )
        try:
            self.firebase_service = firebase_admin.get_app()
        except ValueError as e:
            self.firebase_service = firebase_admin.initialize_app(
                self.service_credentials
            )

        if self.session_name not in st.session_state:
            st.session_state[self.session_name] = None
        if "authentication_token" not in st.session_state:
            st.session_state["authentication_token"] = None
        if "login_message" not in st.session_state:
            st.session_state["login_message"] = None
        if self.authenticator_name not in st.session_state:
            st.session_state[self.authenticator_name] = self
        st.session_state[self.authenticator_name] = self

        self.user_tz = None
        self.cookie_manager = get_cookie_manager()
        if self.debug:
            log.debug(f"Initialized Authenticator {hex(id(self))}")

    def delayed_init(self):
        """
        Used to delay initialization of streamlit objects so this class can be cached
        """
        if self.session_name not in st.session_state:
            st.session_state[self.session_name] = None

        if "authentication_token" not in st.session_state:
            st.session_state["authentication_token"] = None

        if "login_message" not in st.session_state:
            st.session_state["login_message"] = None

        st.session_state[self.authenticator_name] = self

        if not self.user_tz:
            self.user_tz = get_user_timezone(default_tz="Australia/Sydney")

        # if not self.cookie_manager:
        self.cookie_manager.delayed_init()

    @property
    def service_auth(self) -> Optional[ModuleType]:
        user = self.current_user
        if not user or not user.is_admin:
            return None
        else:
            return service_auth

    @property
    def current_user(self) -> Optional[UserInherited]:
        """
        Returns a User interface class instance if there is an active login
        """
        current_user = None
        if self.last_user:
            current_user = self.last_user
        elif (
            self.session_name in st.session_state
            and st.session_state[self.session_name]
        ):
            current_user = st.session_state[self.session_name]

        self.last_user = current_user
        return current_user

    @property
    def auth_cookie(self) -> Optional[str]:
        """
        Gets the auth cookie

        :returns: Returns the auth cookie or None if it doesnt exist
        """
        if (
            "authentication_token" in st.session_state
            and st.session_state["authentication_token"]
        ):
            auth_cookie = st.session_state["authentication_token"]
            return auth_cookie

        cookie = None
        headers = _get_websocket_headers()
        if not headers:
            return cookie

        cookie_header = headers.get("Cookie", None) or headers.get("cookie", None)
        cookie_reader = SimpleCookie(cookie_header)

        cookie = cookie_reader.get(self.cookie_name, None)
        if cookie is not None and cookie.value:
            cookie = cookie.value
        elif self.attempt_fallback_cookies:
            # Try cookie manager component instead of websocket headers
            cookie = self.cookie_manager.get(self.cookie_name)
            cookie = cookie or None  # cookie_manager.get returns None or {}

        return cookie

    def _revoke_auth(
        self, error: Optional[AuthException] = None, disabled: bool = False
    ) -> bool:
        """
        Deletes the auth cookie and revokes any firebase tokens

        :param Optional[AuthException] error: Optional associated error message for the revoke
        :param bool disabled: Optional flag to block execution of this routine. Default is False.

        :returns bool: Always returns False unless an unhandled error is raised
        """
        if disabled:
            log.warning("Authenticator._revoke_auth() disabled")
            return

        msg = error.message if error else ""
        error_type = error.firebase_error if error else ""
        user = self.current_user
        user_display = user.localId if user else "no current_user"
        log.info(f"Revoking cookie for {user_display}: {msg} {error_type}")
        self.cookie_manager.delete(self.cookie_name)

        # If the revokation wasn't from an already expired token, revoke current refresh tokens for user
        if self.last_user and (
            not error or (error and error.firebase_error != "TOKEN_EXPIRED")
        ):
            try:
                service_auth.revoke_refresh_tokens(self.last_user.localId)
                log.info(f"Firebase tokens revoked for {self.last_user.localId}")
            except service_auth.UserNotFoundError:
                log.warning("Couldn't invalidate session. UserNotFoundError")

        # Delete cookie and reset vars
        st.session_state[self.session_name] = None
        st.session_state["authentication_token"] = None
        self.last_user = None
        self.logged_out = True
        self.set_form("login")
        log.warning(f"Logged out {user_display}")
        self.cookie_manager.delete(self.cookie_name)
        # self.cookie_manager.set(self.cookie_name, "", datetime.fromtimestamp(0))

        return False

    @property
    def auth_status(self) -> bool:
        """
        First checks for any valid authentication cookies,
        then returns the current authentication status,
        True if a user is Authenticated, or else False.

        :returns bool: The authentication status
        """
        error = None
        status = False

        if (
            self.current_user
            and st.session_state[self.session_name]
            and st.session_state["authentication_token"]
        ):
            return True

        status, error, session_cookie, decoded_claims = self._check_cookie()
        if error:
            if "Invalid Session Cookie" in str(error) and not self.logged_out:
                log.error(error)
                self._revoke_auth(error)

            return False

        user = None
        if status:
            if (
                self.session_name in st.session_state
                and st.session_state[self.session_name]
            ):
                user = st.session_state[self.session_name]
            if not user:
                user = self._initialize_user_session(session_cookie, decoded_claims)
                if user:
                    self.set_form(None)
                else:
                    error = AuthException("Unable to initialize user")
                    status = None
        else:
            if not self.current_form:
                self.set_form("login")
        self.logged_out = not status
        return status

    def _check_cookie(self) -> tuple[bool, AuthException, str, dict]:
        """
        Looks for a cookie named self.cookie_name and checks its validity
        If it is valid, it returns the session cookie along with decoded claims

        :returns tuple[bool, AuthException, dict]:
            (status, error, session_cookie, decoded_claims)
            status is True if the claims are validated
            error will have any errors
            session_cookie and it's decoded_claims will be set if status is True
        """
        status = False
        error = None
        session_cookie = None
        decoded_claims = None

        if self.logged_out is True:
            return (status, error, session_cookie, decoded_claims)

        session_cookie = self.auth_cookie
        if not session_cookie:
            error = AuthException("No session cookie found")
            return (status, error, session_cookie, decoded_claims)

        try:
            decoded_claims = service_auth.verify_session_cookie(
                session_cookie, check_revoked=True
            )
            if decoded_claims and "user_id" in decoded_claims:
                status = True
                error = None
        except service_auth.InvalidSessionCookieError as e:
            error = AuthException("Invalid Session Cookie")
            return (status, error, session_cookie, decoded_claims)

        if error:
            log.error(f"Unexpected failure in when checking cookie {error}")

        return (status, error, session_cookie, decoded_claims)

    def _initialize_user_session(
        self,
        session_cookie: str,
        decoded_claims: dict,
        login_data: Optional[dict] = None,
        set_cookie_expires_in: Optional[timedelta] = None,
    ) -> Optional[UserInherited]:
        """
        Initialize the users session state after login or reading cookie.
        The cookie and data must be validated before this function.

        :param user: Instance of self.user_class
        :param str session_cookie: the users session cookie
        :param dict decoded_claims: the decoded claims from when verifying login idToken or session cookie
        :param Optional[dict] login_data:
            Account data with idtoken and refresh token returned from pyrebase login handler.
        :param Optional[timedelta] set_cookie_expires_in:
            If you want to write the session to the cookie as well, provide this (it's expiry from now in users tz)

        :returns Optional[UserInherited]:
            Returns the User created
        """
        user = None
        if "user_id" in decoded_claims:
            user_id = decoded_claims["user_id"]
        else:
            user_id = login_data["idToken"]

        user_record = service_auth.get_user(user_id)
        user_record_cleaned = user_record._data
        user_record_cleaned["passwordHash"] = ""
        auth_data = {
            "session_cookie": session_cookie,
            "decoded_claims": decoded_claims,
            "user_record": user_record_cleaned,
        }
        if not login_data:
            login_data = {}
        user = self.user_class(self, auth_data=auth_data, login_data=login_data)

        if user.is_admin:
            service_auth.set_custom_user_claims(user.localId, {"admin": True})
        if user.is_developer:
            service_auth.set_custom_user_claims(user.localId, {"devops": True})
        self.last_user = user

        st.session_state["authentication_token"] = session_cookie
        st.session_state[self.session_name] = user

        if set_cookie_expires_in:
            cookie_expiry_time = (
                datetime.now().astimezone(tz=pytz.timezone(self.user_tz))
                + set_cookie_expires_in
            )
            self.cookie_manager.set(
                self.cookie_name, session_cookie, expires_at=cookie_expiry_time
            )
        log.success(
            f"Initialized session{' and set cookie' if set_cookie_expires_in else ''} {user}"
        )
        return user

    def _create_session(
        self, login_data: dict, expires_in: Optional[timedelta] = None
    ) -> Tuple[Optional[UserInherited], Optional[LoginError]]:
        """
        Creates a User from firebase token data and instantiates a session for them.
        This is used on login.

        :param dict login_data:
            The firebase data to construct the User for the session from - used on login
        :param Optional[timedelta] expires_in:
            a timedelta for how long until the session cookie expires,
            defaults to current time (in user tz) plus self.session_expiry_seconds

        :returns tuple: (user, error) - A user if one was created, and any errors in the operation
        """
        user = None
        error = None
        if not expires_in:
            expires_in = timedelta(seconds=self.session_expiry_seconds)

        id_token = login_data["idToken"]
        try:
            decoded_claims = service_auth.verify_id_token(id_token)
            # Only process if the user signed in within the last 5 minutes.
            if time.time() - decoded_claims["auth_time"] < 5 * 60:
                session_cookie = service_auth.create_session_cookie(
                    id_token, expires_in=expires_in
                )
                user = self._initialize_user_session(
                    session_cookie,
                    decoded_claims,
                    login_data,
                    set_cookie_expires_in=expires_in,
                )
                if not user:
                    error = LoginError("Please try again.")
            else:
                error = LoginError("Expired credentials")
        except (
            service_auth.InvalidIdTokenError,
            service_auth.ExpiredIdTokenError,
            service_auth.RevokedIdTokenError,
            service_auth.CertificateFetchError,
            service_auth.UserDisabledError,
        ):
            error = LoginError("Expired credentials")
        except firebase_admin.exceptions.FirebaseError:
            error = LoginError("Please try again.")

        return (user, error)

    def _check_credentials(
        self, email: str, password: str
    ) -> Tuple[Optional[dict], Optional[LoginError], Optional[dict], bool]:
        """
        Checks the validity of the entered credentials, and if the accounts email is verified.

        :returns tuple: (res, error, login_data, validated)
            res and error are responses and errors from firebase
            login_data and boolean validated are set if the credentials pass
        """
        res = None
        error = None
        login_data = None
        validated = False
        signin_errors = {
            "INVALID_EMAIL": "Invalid email format",
            "EMAIL_NOT_FOUND": "Email address is not registered",
            "INVALID_PASSWORD": "Invalid password",
        }
        res, error = handle_firebase_action(
            auth.sign_in_with_email_and_password,
            LoginError,
            signin_errors,
            email,
            password,
        )
        account_info = None
        user_refresh = None
        if not error:
            refresh_errors = {
                "TOKEN_EXPIRED": "Too many recent sessions. Please try again later."
            }
            user_refresh, error = handle_firebase_action(
                auth.refresh, LoginError, refresh_errors, res["refreshToken"]
            )
            if not error:
                res = {**res, **user_refresh}

        if not error:
            account_info, error = handle_firebase_action(
                auth.get_account_info, LoginError, False, res["idToken"]
            )

        if not error:
            if (
                self.require_email_verification
                and account_info
                and not account_info["users"][0]["emailVerified"]
            ):
                error = LoginError("Please verify your email address before logging in")
                st.button(
                    "Resend verification email",
                    on_click=self._resend_verification,
                    args=(res["idToken"],),
                )

        if res and not error:
            if account_info:
                login_data = res
                account_info["users"][0]["passwordHash"] = ""
                login_data["account_info"] = account_info
                validated = True
            else:
                error = LoginError("Error retrieving account info")

        return (res, error, login_data, validated)

    def _resend_verification(
        self, user_id_token: str
    ) -> Tuple[Optional[dict], Optional[AuthException]]:
        """
        Resends the firebase auth verification email for a user

        :param user_id_token: The authenticated firebase auth token for the user with an unverified email address

        :returns: tuple containing the firebase response and any errors
        """
        res = None
        error = None
        resend_errors = {
            "TOO_MANY_ATTEMPTS_TRY_LATER": "Too many attempts, try again later."
        }
        res, error = handle_firebase_action(
            auth.send_email_verification, AuthException, resend_errors, user_id_token
        )

        if error:
            # st.error(error)
            st.session_state["login_message"] = error
        else:
            # st.info("Verification email sent")
            st.session_state[
                "login_message"
            ] = "Verification email sent, please check your inbox."

        return (res, error)

    def set_form(self, form_name: Optional[str]) -> None:
        """
        Sets a variable to keep track of which form is currently displayed.

        "login" = Login form
        "register" = Registration form
        "reset_password" = Reset password form
        """
        self.current_form = form_name

    def logout(
        self,
        button_name: str,
        button_location: Union[DeltaGenerator, ModuleType] = st.sidebar,
        disabled: bool = False,
    ) -> bool:
        """
        Creates a logout button.

        :param str button_name: The rendered name of the logout button.
        :param Union[DeltaGenerator, ModuleType] button_location: The streamlit container to place the button. Either global `st` or a st container object such as st.sidebar.
        :param bool disabled: Passed to the streamlit button disabled kwarg, to optional disable the button in some circumstances. Default is False.

        :returns bool: Returns True when clicked
        """

        if not isinstance(button_location, DeltaGenerator) and button_location != st:
            raise ValueError(
                "Location must be a streamlit DeltaGenerator (st container object such as st.sidebar) or the global st module."
            )

        if button_location.button(
            button_name,
            disabled=disabled,
            on_click=self._revoke_auth,
            args=(
                AuthException("Logout requested"),
                disabled,
            ),
            key=f"authenticator_logout_btn_{time.time()}",
        ):
            return True

        return False

    def login(
        self,
        form_name: str,
        form_location: Union[DeltaGenerator, ModuleType] = st,
        success_callback: Optional[Callable] = None,
        cb_args: Optional[tuple] = None,
        cb_kwargs: Optional[dict] = None,
    ) -> Tuple[Optional[UserInherited], dict, Optional[LoginError]]:
        """
        Creates a login widget.

        :param str form_name: The rendered name of the login form.
        :param Union[DeltaGenerator, ModuleType] form_location: The streamlit container to place the form. Either global `st` or a st container object such as st.sidebar.
        :param Callable success_callback:
            Optional function to call if a user is created without error.
            Will be passed a kwargs login_return which will be the same tuple returned from this login function.
        :param cb_args: Extra args for success_callback
        :param cb_kwargs: Extra keyword args for success_callback

        :returns: 3 element tuple containing the a User class if one was created, response from the firebase calls, and/or any errors
        """
        if not isinstance(form_location, DeltaGenerator) and form_location != st:
            raise ValueError(
                "Location must be a streamlit DeltaGenerator (st container object such as st.sidebar) or the global st module."
            )

        res = None
        user = None
        error = None

        if cb_args is None:
            cb_args = ()
        if cb_kwargs is None:
            cb_kwargs = {}

        auth_status = self.auth_status
        if not auth_status:
            login_form = form_location.form("Login")
            login_form.subheader(form_name)

            if st.session_state["login_message"]:
                if isinstance(st.session_state["login_message"], Exception):
                    login_form.error(st.session_state["login_message"])
                else:
                    login_form.info(st.session_state["login_message"])
                st.session_state["login_message"] = None

            email = login_form.text_input("Email").lower()
            password = login_form.text_input("Password", type="password")

            if login_form.form_submit_button("Login"):
                if len(email) == 0:
                    error = LoginError("Please enter an email")
                elif len(password) == 0:
                    error = LoginError("Please enter a password")
                else:
                    res, error, login_data, validated = self._check_credentials(
                        email, password
                    )
                    if validated and login_data:
                        user, error = self._create_session(login_data)

                    if (
                        user
                        and not error
                        and success_callback
                        and callable(success_callback)
                    ):
                        success_callback(
                            *cb_args,
                            login_return=(
                                user,
                                res,
                                error,
                            ),
                            **cb_kwargs,
                        )

            login_form.form_submit_button(
                "Register", on_click=self.set_form, args=("register",)
            )
            login_form.form_submit_button(
                "Forgot password?", on_click=self.set_form, args=("reset_password",)
            )

        return (user, res, error)

    def register_user(
        self,
        form_name: str,
        form_location: Union[DeltaGenerator, ModuleType] = st,
        terms_link: Optional[str] = None,
    ) -> Tuple[Optional[dict], Optional[RegisterError]]:
        """
        Creates a user registration widget

        :param str form_name: The rendered name of the password reset form.
        :param Union[DeltaGenerator, ModuleType] form_location: The streamlit container to place the form. Either global `st` or a st container object such as st.sidebar.

        :returns: tuple containing the firebase response and any errors
        """
        if not isinstance(form_location, DeltaGenerator) and form_location != st:
            raise ValueError(
                "Location must be a streamlit DeltaGenerator (st container object such as st.sidebar) or the global st module."
            )

        register_user_form = form_location.form("Register user")
        register_user_form.subheader(form_name)
        new_email = register_user_form.text_input("Email")
        new_name = register_user_form.text_input("Name")
        new_password = register_user_form.text_input("Password", type="password")
        new_password_repeat = register_user_form.text_input(
            "Repeat password", type="password"
        )

        if terms_link is not None:
            terms_agreed = register_user_form.checkbox(
                f"By registering I agree to the terms and conditions [here]({terms_link}).",
            )
        else:
            terms_agreed = True

        user = None
        error = None
        if register_user_form.form_submit_button("Register"):
            if not terms_agreed:
                error = RegisterError(
                    "Please read and agree to the usage terms and conditions."
                )
            elif len(new_email) > 0 and len(new_name) > 0 and len(new_password) > 0:
                if new_password == new_password_repeat:
                    register_errors = {
                        "INVALID_EMAIL": "Invalid email format",
                        "EMAIL_EXISTS": "Email is already registered",
                        "WEAK_PASSWORD": "",
                    }
                    user, error = handle_firebase_action(
                        auth.create_user_with_email_and_password,
                        RegisterError,
                        register_errors,
                        new_email,
                        new_password,
                    )
                    # user = service_auth.create_user(
                    # email='user@example.com',
                    # email_verified=False,
                    # phone_number='+15555550100',
                    # password='secretPassword',
                    # display_name='John Doe',
                    # photo_url='http://www.example.com/12345678/photo.png',
                    # disabled=False)

                    if not error:
                        res1, error = handle_firebase_action(
                            auth.update_profile,
                            RegisterError,
                            False,
                            user["idToken"],
                            display_name=new_name,
                        )

                    if not error:
                        res2, error = handle_firebase_action(
                            auth.send_email_verification,
                            RegisterError,
                            False,
                            user["idToken"],
                        )
                else:
                    error = RegisterError("Passwords do not match")
            else:
                error = RegisterError(
                    "Please enter an email, username, name, and password"
                )

        register_user_form.form_submit_button(
            "Cancel", on_click=self.set_form, args=(None,)
        )

        if user and self.require_email_verification:
            verify_msg = """<p style="padding: 14px;">Registration successful. Please verify your email address, then you can <a href="/" target="_self">log in.</a></p>"""
            register_user_form.markdown(
                custom_html.custom_el(verify_msg, classes=custom_html.st_info_classes),
                unsafe_allow_html=True,
            )

        return (user, error)

    def reset_password(
        self, form_name: str, form_location: Union[DeltaGenerator, ModuleType] = st
    ) -> Tuple[Optional[dict], Optional[ResetError]]:
        """
        Creates a reset password via email widget.

        :param str form_name: The rendered name of the reset password form.
        :param Union[DeltaGenerator, ModuleType] form_location: The streamlit container to place the form. Either global `st` or a st container object such as st.sidebar.

        :returns: tuple containing the firebase response and any errors
        """
        if not isinstance(form_location, DeltaGenerator) and form_location != st:
            raise ValueError(
                "Location must be a streamlit DeltaGenerator (st container object such as st.sidebar) or the global st module."
            )

        reset_password_form = form_location.form("Reset password")
        reset_password_form.subheader(form_name)
        email = reset_password_form.text_input("Email").lower()

        res = None
        error = None
        if reset_password_form.form_submit_button("Submit"):
            if len(email) > 0:
                reset_errors = {
                    "EMAIL_NOT_FOUND": "Email address is not registered",
                    "INVALID_EMAIL": "Invalid email format",
                }
                res, error = handle_firebase_action(
                    auth.send_password_reset_email, ResetError, reset_errors, email
                )
            else:
                error = ResetError("Please enter an email")

        reset_password_form.form_submit_button(
            "Cancel", on_click=self.set_form, args=(None,)
        )

        return (res, error)

    def __repr__(self) -> str:
        return repr_(
            self,
            [
                "cookie_key",
                "service_credentials",
                "auth",
                "storage",
                "firebase_service",
            ],
        )


# @st.cache(allow_output_mutation=True, show_spinner=False, hash_funcs={"_thread.RLock": lambda _: None, "weakref.ReferenceType": lambda _: None})
def get_auth(
    cookie_name: str,
    user_class: Optional[UserInherited] = None,
    authenticator_name="authenticator",
    **kwargs,
) -> Authenticator:
    """
    See Authenticator for params.
    """
    if authenticator_name in st.session_state and st.session_state[authenticator_name]:
        return st.session_state[authenticator_name]

    cookie_key = st.secrets["authenticator"]["cookie_key"]
    admin_ids = st.secrets["authenticator"]["admin_ids"]
    developer_ids = st.secrets["authenticator"]["developer_ids"]
    return Authenticator(
        cookie_name,
        cookie_key,
        authenticator_name=authenticator_name,
        user_class=user_class,
        admin_ids=admin_ids,
        developer_ids=developer_ids,
        **kwargs,
    )



import json
import requests


# Proxy for handling pyrebase/firebase errors
def handle_firebase_action(
    action_function, exception_type, error_dict, *fn_args, **fn_kwargs
):
    """
    Handler for pyrebase/firebase actions.
    """
    error = None
    result = None
    try:
        # print("Args", fn_args, fn_kwargs)
        result = action_function(*fn_args, **fn_kwargs)
    except requests.HTTPError as e:
        # Pyrebase/4 is intercepting and raising an incorrectly initialized requests.HTTPError,
        # so the response has to be extracted from builtin BaseException args rather than e.response
        response_dict = json.loads(e.args[0].response.text.replace('"', '"'))
        error_type = None
        try:
            error_type = response_dict["error"]["message"]
        except:
            try:
                error_type = response_dict["error"]["errors"][0]["message"]
            except:
                pass

        if e.args[0].response.status_code == 400:
            if not error_dict:
                error = exception_type(
                    f"Unknown {exception_type.__name__} while trying to {action_function.__name__} ({error_type})",
                    error_type,
                    e.args[0],
                )
            else:
                if error_type in error_dict:
                    error = exception_type(
                        error_dict[error_type], error_type, e.args[0]
                    )
                else:
                    for key in error_dict:
                        if key in error_type:
                            message = error_dict[key] + " " + error_type.split(" : ")[1]
                            error = exception_type(message, error_type, e.args[0])
                if not error:
                    error = exception_type(
                        f"Unknown {exception_type.__name__} while trying to {action_function.__name__} ({error_type})",
                        error_type,
                        e.args[0],
                    )
        else:
            error = exception_type(
                f"Storage network error, please try again later ({e.args[0].response.status_code}) ({error_type})",
                error_type,
                e.args[0],
            )
            print(e.args[0])
        if not error:
            raise
    except requests.ConnectionError as e:
        error = exception_type(
            "Storage connection error, please try again later", f"SERVER_{e.args[0]}", e
        )

    return (result, error)



import time
import streamlit as st
from urllib.parse import quote, unquote
from typing import Tuple, Optional, Callable
from streamlitextras.webutils import stxs_javascript
from streamlitextras.logger import log
from streamlitextras.utils import repr_


class Router:
    """
    Page router for streamlit.

    :param dict[str, Callable] routes:
        Dictionary mapping of routes to their page functions, in the format {page_name: page_function}
    :param Optional[Callable] preroute:
        Optional callable page function that will be executed before each page function
    :param Optional[list] dependencies:
        Optional dict to pass as kwargs to every page_function call
    :param Optional[Callable] postroute:
        Optional callable page function that will be executed after each page function
    """

    def __init__(
        self,
        routes: dict[str, Callable],
        preroute: Optional[Callable] = None,
        dependencies: Optional[dict] = None,
        postroute: Optional[Callable] = None,
        debug: bool = False,
    ):
        self.routes = routes
        self.preroute = preroute
        self.dependencies = dependencies
        self.postroute = postroute
        self.page_names = list(self.routes.keys())
        self.debug = debug
        if self.debug:
            log.debug(f"Initialized router {hex(id(self))}")

    def delayed_init(self):
        """
        Used to delay initialization of streamlit objects so this class can be cached
        """
        st.session_state["router"] = self

    @property
    def default_page(self):
        """
        Returns the default page. Currently the first in self.routes
        """
        return self.page_names[0]

    @property
    def current_page(self):
        page_name, page_state = self.current_page_data()
        return page_name

    @property
    def current_state(self):
        page_name, page_state = self.current_page_data()
        return page_state

    @property
    def query_params(self):
        """
        Uses streamlit to fetch the current URL query parameters,
        removing the current page param and state value.
        """
        params = st.experimental_get_query_params()
        params = {k: v[0] for k, v in params.items() if k not in self.routes}
        return params

    @property
    def query_params_all(self):
        """
        Uses streamlit to fetch the current URL query parameters,
        including the current page param and state value.
        """
        params = st.experimental_get_query_params()
        return params

    def current_page_data(self) -> Tuple:
        """
        Returns the current page name and page from the query string
        """
        query_params = st.experimental_get_query_params()
        page_name = None
        page_state = None

        for q_page_name, q_page_state in query_params.items():
            if q_page_name in self.page_names:
                page_state = q_page_state
                page_name = q_page_name
                break

        if not page_name:
            page_name = self.default_page

        if page_state and page_state[0] != "None":
            page_state = unquote(page_state[0])
        else:
            page_state = None

        return (page_name, page_state)

    def show_route_view(
        self,
        force_page_name: Optional[str] = None,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
        redirect_page_names: Optional[list[str]] = None,
    ):
        """
        Checks the query params and routes to the requested page,
        or routes to force_page_name directly, not setting any query params

        :param force_page_name:
            If provided will route to this page via its function in self.routes,
            bypassing query string params.
        :param args:
            Tuple of args to pass to the page func for force_page_name
        :param kwargs:
            Dict of kwargs to pass to the page func for force_page_name
            They will take precedence over and be merged with self.dependencies
        :param redirect_page_names:
            If page name from query string is in this list, will be redirected to the default route instead.
            Useful for conditional redirection in authentication etc.
        """
        if force_page_name:
            page_name = force_page_name
        else:
            page_name, page_state = self.current_page_data()
            args = (page_state,)

        redirect = False
        if redirect_page_names and page_name in redirect_page_names:
            redirect = True
            page_name = self.default_page
            kwargs = {}
            args = ()

        page_func = self.routes[page_name]

        if type(kwargs) != dict:
            kwargs = {}

        if self.dependencies and type(self.dependencies) is dict:
            kwargs = {**self.dependencies, **kwargs}

        if args is None or args in [("",), ("~",), ("None",), "", "~", "None"]:
            args = ()

        if self.preroute and callable(self.preroute):
            # log.debug(f"Running preroute {self.preroute}")
            self.preroute(*args, **kwargs)

        if self.debug:
            log.info(f"Routing to {page_name} {redirect}")

        if redirect is True:
            st.experimental_set_query_params(**{page_name: "~"})
            st.experimental_rerun()

        if callable(page_func):
            # log.debug(f"Calling page_func {page_func}")
            page_func(*args, **kwargs)

        if self.postroute and callable(self.postroute):
            # log.debug(f"Running postroute {self.postroute}")
            self.postroute(*args, **kwargs)

        if len(args) == 0:
            stxs_javascript(
                f"""window.history.pushState({{}}, "", "/?{page_name}=~");"""
            )

    def route(
        self,
        page_name: str = None,
        page_state: Optional[str] = None,
        additional_params: Optional[dict] = None,
        rerun_st: bool = False,
    ):
        """
        Routes to a page.
        First found query string matching a page key in self.routers is routed too.
        Query string value can be set to page data

        :param Optional[str] page_name:
            The key for the page in self.routes - query param key will be set the same
            If it is None first page in self.routes will be used, and no query params will be set (redirect to /)
        :param Optional[str] page_state:
            Optional string to include as page state, will be urlencoded/urldecoded
        :param Optional[dict] additional_params:
            Optional dict to be set as query parameters using st.experimental_set_query_params
            If you use the page name as one of the keys, behaviour is overriding and may be experimental
        :param bool rerun_st:
            Whether to call st.experimental_rerun() - not needed if calling this from a st callback
        """
        if additional_params is None:
            additional_params = {}

        query_params = {}
        page_state = quote(page_state) if page_state else "~"
        if page_name:
            query_params = {page_name: page_state}
        else:
            page_name = self.default_page
            query_params = {page_name: page_state}

        if self.debug:
            log.info(f"Setting query params {query_params}")

        st.experimental_set_query_params(**{**query_params, **additional_params})
        if rerun_st is True:
            if self.debug:
                log.debug("rerun_st is True")
            time.sleep(0.1)
            st.experimental_rerun()

    def __repr__(self) -> str:
        return repr_(self, ["routes"])


router_hash_funcs = {
    "_thread.RLock": lambda _: None,
    "builtins.method": lambda _: None,
    "builtins.property": lambda _: None,
    "builtins.function": lambda _: None,
    "_cffi_backend.__CDataGCP": lambda _: None,
    "google.cloud.storage.client.Client": lambda _: None,
    "streamlit.delta_generator.DeltaGenerator": lambda _: None,
}


# @st.cache(allow_output_mutation=True, show_spinner=False, hash_funcs=router_hash_funcs)
def get_router(
    routes: dict[str, Callable],
    preroute: Optional[Callable] = None,
    dependencies: Optional[dict] = None,
    postroute: Optional[Callable] = None,
) -> Router:
    """
    See Router for params.
    """
    if "router" in st.session_state and (
        router := st.session_state.get("router", None)
    ):
        router: Router
        router.routes = routes or router.routes
        router.preroute = preroute or router.preroute
        router.postroute = postroute or router.postroute
        router.dependencies = dependencies or router.dependencies
        return router
    return Router(routes, preroute, dependencies, postroute)



# Exceptions for the firebase auth helper in autenticator.py
from typing import Union, Optional
from requests import HTTPError


class AuthException(Exception):
    """
    Base exception for Authenticator class
    """

    def __init__(
        self,
        message: str,
        firebase_error: Optional[str] = None,
        requests_exception: Optional[Union[HTTPError, Exception]] = None,
    ):
        """
        :param str message: message to be displayed to the user
        :param Optional[str] firebase_error: firebase error message enum e.g INVALID_PASSWORD
        :param Union[HTTPError, Exception] requests_exception: requests.HTTPError with .response and .request attributes (or other associated exception)
        """
        self.message = message
        self.firebase_error = firebase_error
        self.requests_exception = requests_exception


class LoginError(AuthException):
    """Exceptions raised for the login user widget."""


class RegisterError(AuthException):
    """Exceptions raised for the register user widget."""


class ResetError(AuthException):
    """Exceptions raised for the rest password widget."""


class UpdateError(AuthException):
    """Exceptions raised for the update user details widget."""



# from .router import Router
# from .cookiemanager import CookieManager
# from .authenticator import Authenticator
# from .authenticator.user import User
# from .authenticator.exceptions import AuthException, LoginError, RegisterError, ResetError, UpdateError

__version__ = "0.2.39"



import os
import sys
import ast
import loguru
import dateutil.parser
from loguru._logger import Logger
import streamlit as st
from streamlitextras.utils import repr_

dev_emulation = os.environ.get("DEV_EMULATION", False)

_LOGGER = loguru.logger
log_folder = "logs"
log_filename = "streamlit.log"

module_filter = ""
default_format = "{time} | {level: <8} | {name}:{module}:{function}:{file}:{line} | {message} | {extra}"
detailed_format = "{time} | {level: <8} | {name}:{module}:{function}:{file.path}:{line} | {message} | {extra} | {exception} | {process.name}:{process} {thread.name}:{thread}"

colour_format = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
    "<level>{level: <8}</level> | "
    "<cyan>{name}</cyan>:<cyan>{module}</cyan>:<cyan>{function}</cyan>:<cyan>{file}</cyan>:<cyan>{line}</cyan> | {{extra[user]}} | "
    "\n<level>{message}</level>"
)

colour_detailed_format = (
    "<red>{time:YYYY-MM-DD HH:mm:ss.SSS}</red> | "
    "<level>{level: <8}</level> | "
    "<cyan>{name}</cyan>:<cyan>{module}</cyan>:<cyan>{function}</cyan>:<cyan>{file.path}</cyan>:<cyan>{line}</cyan> | "
    "\n<level>{message} | {extra[user]}</level> | "
    "<level>{exception}</level> | "
    "<level>{process.name}:{process} @ {thread.name}:{thread}</level>"
)
if dev_emulation:
    colour_format = colour_format.replace("file", "file.path")

handlers = [
    {
        "sink": os.path.join(log_folder, log_filename),
        "rotation": "250 MB",
        "enqueue": True,
        "format": default_format,
        "filter": module_filter,
        "level": "INFO",
        "catch": True,
    },
    {
        "sink": os.path.join(log_folder, "debug_" + log_filename),
        "rotation": "250 MB",
        "enqueue": True,
        "format": detailed_format,
        "filter": module_filter,
        "level": "TRACE",
        "backtrace": True,
        "diagnose": True,
        "catch": True,
    },
    # Errors will be logged to sys.stdout
    # {"sink": sys.stderr,
    # "format": colour_detailed_format,
    # "filter": module_filter,
    # "level": "ERROR",
    # "backtrace": True,
    # "colorize": True},
    {
        "sink": sys.stdout,
        "format": colour_format,
        "filter": module_filter,
        "level": "DEBUG",
        "colorize": True,
    },
]


def process_log_line(log_line: str):
    """
    Process a log line formatted from this module into a dictionary of its sections.

    :param str log_line: The log line to process
    """
    if not log_line:
        return None
    time, level, namespace, message, extra, exception, exec = log_line.split(" | ")
    level = level.strip()
    name, module, function_name, file_path, line = namespace.split(":")

    try:
        process, thread = exec.split(" @ ")
    except:
        process = exec.split(" ")[0]
        thread = " ".join(exec.split(" ")[1:])
    process_name, process_id = process.split(":")
    thread_name, thread_id = thread.split(":")

    log_obj = {
        "log_line": log_line,
        "level": level,
        "time": dateutil.parser.isoparse(time),
        "message": message,
        "extra": ast.literal_eval(extra),
        "exception": exception,
        "namespace": namespace,
        "name": name,
        "module": module,
        "function_name": function_name,
        "file_path": file_path,
        "line": line,
        "process_name": process_name,
        "process_id": process_id,
        "thread_name": thread_name,
        "thread_id": thread_id,
    }
    return log_obj


def session_state_str_dict():
    """
    Generate a dict from st.session_state to be stored with every log line
    """
    interface = {"user": None, "session_state": None}
    extra = {
        "user": repr(st.session_state.get("user", None)),
        "session_state": {k: f"{v}" for k, v in st.session_state.to_dict().items()},
    }
    merged = {**interface, **extra}
    return merged


def bind_log(extras={}) -> Logger:
    """
    Bind the logger to the session state dictionary
    """
    merged = {**session_state_str_dict(), **extras}
    log = _LOGGER.bind(**merged)
    return log


log: Logger


def __getattr__(name):
    if name == "log":
        log = bind_log()
        return log
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


sinks = _LOGGER.configure(handlers=handlers, activation=[("", True)])



import os
import time
import streamlit as st
import streamlit.components.v1 as components
from streamlitextras.logger import log
from streamlitextras.utils import repr_

absolute_path = os.path.dirname(os.path.abspath(__file__))
build_path = os.path.join(absolute_path, "frontend/build")
_component_func = components.declare_component("cookie_manager", path=build_path)


class CookieManager:
    """
    This is a streamlit component class to manage cookies.

    It uses a thin component instance wrapper around the universal-cookie library.
    """

    def __init__(self, debug: bool = False):
        if "cookie_manager" in st.session_state and st.session_state["cookie_manager"]:
            self.cookies = st.session_state["cookie_manager"].cookies.copy()
        else:
            self.cookies = {}
        st.session_state["cookie_manager"] = self
        self.debug = debug
        if self.debug:
            log.debug(f"Initialized Cookie Manager {hex(id(self))}")

    def delayed_init(self):
        """
        Used to delay initialization of streamlit objects so this class can be cached
        """
        st.session_state["cookie_manager"] = self

    def cookie_manager(self, *args, **kwargs):
        time.sleep(0.1)
        try:
            result = _component_func(**kwargs)
        except st.errors.DuplicateWidgetID:
            kwargs["key"] = f"""{kwargs["key"]}{time.time()}"""
            result = _component_func(**kwargs)
        time.sleep(0.1)  # This ensures we get a result before streamlit redraws
        return result

    def set(
        self,
        name,
        value,
        expires_at=None,
        secure=None,
        path=None,
        same_site=None,
        key="set",
    ):
        """
        Set a cookie with name, value and options.
        Defaults are set in the JS component and listed below

        :param name: The name of the cookie
        :param value: The value of the cookie
        :param expires_at:
            Datetime of when the cookie expires.
            Default is set in the js at 1 day from browser timezone.
        :param secure: Secure flag, default is true.
        :param path: Cookie path, default is /
        :param same_site: Same site attribute, default is "strict"

        :param key: streamlit key used for the component instance

        :returns: True if the operation was successful, else False
        """
        if name is None or name == "":
            return False

        expires_at = expires_at.isoformat() if expires_at else None
        options = {
            "name": name,
            "value": value,
            "expires": expires_at,
            "sameSite": same_site,
        }
        result = self.cookie_manager(
            method="set", options=options, key=key, default=False
        )
        if result:
            self.cookies[name] = result

        return True

    def get(self, name, key="get"):
        """
        Gets a value of a cookie.
        NOTE: The value isn't got directly from the component instance
        This class fills all cookies in self.cookies everytime streamlit instances it on the page

        Returns empty {} dict if no result. Returns None if the operation failed.

        :param name: The name of the cookie to get the value of

        :returns Union[str, dict]:
            Returns the cookie value, it may be deserialized from JSON to a dict
        """
        if not self.cookies:
            self.cookies = self.cookie_manager(method="getAll", key=key, default={})

        if name is None or name == "":
            return None
        result = self.cookies.get(name, None)
        time.sleep(0.4)  # Give component time to render
        return result

    def get_all(self, key="get_all"):
        """
        Get a dict of all the cookies in the browser, and update self.cookies

        :param key: streamlit key used for the component instance

        :returns: A dict of all the cookies
        """
        self.cookies = self.cookie_manager(method="getAll", key=key, default={})
        return self.cookies

    def delete(self, name, key="delete"):
        """
        Delete a cookie from the browser

        :param name: the name of the cookie to delete
        :param key: streamlit key used for the component instance

        :returns: True if the operation was successful, or else False
        """
        if name is None or name == "":
            return False

        result = self.cookie_manager(
            method="delete", options={"name": name}, key=key, default=False
        )
        if result and name in self.cookies:
            del self.cookies[name]
        return result

    def __repr__(self) -> str:
        return repr_(self)


# @st.cache(allow_output_mutation=True, show_spinner=False)
def get_cookie_manager() -> CookieManager:
    if "cookie_manager" in st.session_state and st.session_state["cookie_manager"]:
        return st.session_state["cookie_manager"]
    return CookieManager()



import base64
import hashlib
import binascii
import datetime
import streamlit as st
from streamlit.runtime.uploaded_file_manager import UploadedFile

from google.oauth2 import service_account
from google.cloud import storage
from google.cloud.storage.retry import _should_retry
from google.api_core import retry

from typing import Optional, Iterable, Union, Tuple
from io import BytesIO

# Default project and bucket
default_gproject = st.secrets["gcp_service_account"]["project_id"]
default_bucket = st.secrets["firebase"]["storageBucket"]

# Create API client.
credentials = service_account.Credentials.from_service_account_info(
    dict(st.secrets["gcp_service_account"])
)
client = storage.Client(credentials=credentials)

_INITIAL_DELAY = 1.0  # seconds
_MAXIMUM_DELAY = 4.0
_DELAY_MULTIPLIER = 2.0
_DEADLINE = 60.0
retry_custom = retry.Retry(
    _should_retry, _INITIAL_DELAY, _MAXIMUM_DELAY, _DELAY_MULTIPLIER, _DEADLINE
)


def compute_bytes_md5hash(data: bytes):
    md5hex = hashlib.md5(data).hexdigest()
    encoded_bytes = base64.b64encode(binascii.unhexlify(md5hex))
    sig = encoded_bytes.rstrip(b"\n").decode("UTF-8")
    return sig


def get_buckets(
    project: str = default_gproject,
    prefix: Optional[str] = None,
    page_size: int = 500,
    timeout: int = 8,
    retry: retry.Retry = retry_custom,
    print_names: bool = False,
) -> Iterable[storage.Bucket]:
    buckets = client.list_buckets(
        project=project,
        prefix=prefix,
        page_size=page_size,
        timeout=timeout,
        retry=retry,
    )
    if print_names:
        for bucket in buckets:
            print(bucket)
    return buckets


def get_blobs(
    bucket_or_name: Union[str, storage.Bucket] = default_bucket,
    prefix: Optional[str] = None,
    page_size: int = 500,
    timeout: int = 8,
    retry: retry.Retry = retry_custom,
    print_names: bool = False,
) -> Optional[Iterable[storage.Blob]]:
    """
    Prints a list of all the blobs in the specified storage bucket,
    and returns them as an iterable
    """
    blobs = client.list_blobs(
        bucket_or_name, prefix=prefix, page_size=page_size, timeout=timeout, retry=retry
    )
    bucket_name = bucket_or_name if type(bucket_or_name) == str else bucket_or_name.name
    if print_names:
        for blob in blobs:
            print(blob)

    return blobs


def get_bucket(
    bucket_name: str, timeout: int = 8, retry: retry.Retry = retry_custom
) -> Optional[storage.Bucket]:
    bucket = client.lookup_bucket(bucket_name, timeout=timeout, retry=retry)
    return bucket


def create_bucket(
    bucket_name: str,
    project: str = default_gproject,
    billing_project: str = default_gproject,
    bucket_location: str = "us-east1",
    timeout: int = 8,
    retry: retry.Retry = retry_custom,
) -> storage.bucket.Bucket:
    bucket = client.create_bucket(
        bucket_name,
        location=bucket_location,
        project=project,
        user_project=billing_project,
        timeout=timeout,
        retry=retry,
    )
    return bucket


def upload_blob_data(
    blob_name: str,
    blob_data: Union[UploadedFile, str],
    bucket_name: str = default_bucket,
    content_type: Optional[str] = None,
    timeout: int = 8,
    retry: retry.Retry = retry_custom,
) -> Tuple[storage.Blob, str]:
    bucket = client.bucket(bucket_name)
    blob = bucket.blob(blob_name)

    if blob.exists(client, timeout=timeout, retry=retry):
        return (blob, get_blob_url(blob))

    if type(blob_data) == UploadedFile:
        blob.upload_from_file(
            blob_data, content_type=content_type, timeout=timeout, retry=retry
        )
    elif type(blob_data) == str:
        blob.upload_from_string(
            blob_data, content_type=content_type, timeout=timeout, retry=retry
        )

    return (blob, get_blob_url(blob))


def get_blob_url(blob: storage.Blob):
    url = blob.generate_signed_url(
        expiration=datetime.timedelta(minutes=60), version="v4", method="GET"
    )
    return url


def download_blob_data(
    bucket_name: str, blob_name: str, return_type: str = "bytes"
) -> Union[BytesIO, str]:
    bucket = client.bucket(bucket_name)
    content = None
    if return_type == "bytes":
        content = bucket.blob(blob_name).download_as_bytes()
    elif return_type == "string":
        content = bucket.blob(blob_name).download_as_text(encoding="utf-8")

    return content



import re

package_name = "streamlit-base-extras"
version_init_file = "streamlitextras/__init__.py"

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages


with open(version_init_file, "r") as file:
    regex_version = r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]'
    version = re.search(regex_version, file.read(), re.MULTILINE).group(1)

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name=package_name,
    version=version,
    author="blipk",
    author_email="blipk+@github.com",
    description="Make building with streamlit easier.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/blipk/streamlitextras",
    packages=find_packages(),
    include_package_data=True,
    download_url="https://github.com/blipk/streamlitextras/archive/{}.tar.gz".format(
        version
    ),
    project_urls={
        "Changelog": "https://github.com/blipk/streamlitextras/commits/",
        "Documentation": "https://streamlitextras.readthedocs.io/en/stable/index.html",
    },
    keywords=[
        "streamlitextras",
        "streamlit",
        "router",
        "authenticator",
        "javascript",
        "cookie",
        "thread",
    ],
    license="MIT license",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Browsers",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    install_requires=[
        "streamlit >= 1.23.1",
        "streamlit-javascript == 0.1.5",
        "loguru == 0.6.0",
        "requests",
        "gcloud",
        "pyjwt",
        "firebase",
        "pyrebase4",
        "sseclient",
        "python_jwt",
        "PyCryptodome",
        "requests_toolbelt",
        "firebase-admin",
        "google-cloud-storage",
        "streamlit_javascript",
    ],
    extras_require={
        "dev": [
            # Docs
            "Sphinx==5.2.3 ; python_version>='3.6'",
            "sphinx-autobuild==2021.3.14 ; python_version>='3.6'",
            "sphinx-rtd-theme==1.0.0 ; python_version>='3.6'",
            "docutils==0.16 ; python_version>='3.6'",
            "sphinxcontrib-apidoc ; python_version>='3.6'",
        ]
    },
    python_requires=">=3.5",
)



import time
import html
import uuid
import base64
import streamlit as st
from io import BytesIO
from typing import Union, Optional
from streamlit_javascript import st_javascript
from streamlit.runtime.uploaded_file_manager import UploadedFile


def stxs_javascript(source: str) -> None:
    """
    Runs javascript on the top level context of the page.

    Does this by embedding an iframe that attaches a script tag to its parent.

    :param str source: The script source to embed in the <script></script> element
    """
    div_id = uuid.uuid4()

    st.markdown(
        f"""
    <div style="display:none" id="{div_id}">
        <iframe src="javascript: \
            var script = document.createElement('script'); \
            script.type = 'text/javascript'; \
            script.text = {html.escape(repr(source))}; \
            var thisDiv = window.parent.document.getElementById('{div_id}'); \
            var rootDiv = window.parent.parent.parent.parent.document.getElementById('root'); \
            rootDiv.appendChild(script); \
            thisDiv.parentElement.parentElement.parentElement.style.display = 'none'; \
        "/>
    </div>
    """,
        unsafe_allow_html=True,
    )
    time.sleep(0.1)
    return True


def get_user_timezone(default_tz: Optional[str] = None) -> Optional[str]:
    """
    Uses javascript to get the tz database name for the browser/users timezone

    :param Optional[str] default_tz: value to return if the operation fails. Defaults to UTC

    :returns Optional[str]:
        The tz database name for the timezone,
        or None if the operation fails and default_tz isn't set, or isn't supported by the browser (unlikely)
    """
    if not default_tz:
        default_tz = "Etc/UTC"

    timezone = st_javascript(
        """await (async () => {
            const userTimezone = Intl.DateTimeFormat().resolvedOptions().timeZone
            console.log(userTimezone)
            return userTimezone
        })().then(returnValue => returnValue)"""
    )
    if timezone == 0:  # st_javascript returns 0 for null/undefined
        timezone = default_tz
    return timezone


def scroll_page(x: int = 0, y: int = 0):
    """
    Runs javascript to scroll the streamlit main section
    Defaults to scrolling to the top if no arguments are passed

    :param int x: the x coordinate to scroll to
    :param int y: the y coordinate to scroll to
    """
    stxs_javascript(
        f"""parent.document.querySelector(`section[class*="main"`).scroll({x}, {y})"""
    )


def bytes_to_data_uri(
    byteslike_object: Union[BytesIO, UploadedFile, bytes],
    mime_type: Optional[str] = None,
) -> str:
    """
    Creates a data URI from a bytesIO object

    :param Union[BytesIO, UploadedFile] byteslike_object: BytesIO or bytes or any class that inherits them
    :param str mime_type: The mimetype to set on the data URI
    :returns: The data URI as a string
    """
    data = None
    try:
        data = byteslike_object.getvalue()
    except:
        data = byteslike_object
    if not mime_type:
        mime_type = "application/octet-stream"
    uri = f"data:{mime_type};base64,{base64.b64encode(data).decode()}"
    return uri


def trigger_download(download_uri: str, filename: str):
    """
    Uses javascript and a data URI on a link element to trigger a download for the user

    :param str download_uri: properly formatted data uri to place on link elements href
    :param str filename: filename to be placed on the link elements download attribute
    """
    auto_downloaded = stxs_javascript(
        f"""(async () => {{
            console.log("Creating download link..")
            var link = document.createElement('a');
            link.innerText = ""
            link.href = "{download_uri}";
            link.target = "_blank";
            link.download = "{filename}";
            link.click();
            await new Promise(r => setTimeout(r, 2000));
            //window.open(link.href, "_blank")
        }})();"""
    )
    time.sleep(
        1
    )  # Give the page some time to render the element before the page rerenders
    return auto_downloaded


def convert_millis(milliseconds: int, always_include_hours: bool = False) -> str:
    """
    Convert milliseconds to a string timestamp in the format HH:MM:SS or MM:SS

    :param int milliseconds: The amount of milliseconds to convert to a timestamp
    :param bool always_include_hours: Always include the HH: part of the timestamp even if 00.
    """
    seconds = int((milliseconds / 1000) % 60)
    minutes = int((milliseconds / (1000 * 60)) % 60)
    hours = int((milliseconds / (1000 * 60 * 60)) % 24)
    timestamp_text = ""
    if hours > 0:
        timestamp_text += f"{hours:02d}:{minutes:02d}:{seconds:02d}"
    else:
        if always_include_hours:
            timestamp_text += f"{hours:02d}:"
        timestamp_text += f"{minutes:02d}:{seconds:02d}"

    return timestamp_text
